/*
 *  <Short Description>
 *  Copyright (C) 2024  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/renderer/font_renderer.h>
#include <blt/gfx/renderer/shaders/2d_font.vert>
#include <blt/gfx/renderer/shaders/2d_font.frag>
#include "blt/gfx/shader.h"

#include <utility>

namespace blt::gfx
{
    font_texture_atlas::font_texture_atlas(blt::i32 dimensions): texture_gl2D(dimensions, dimensions, GL_R8)
    {}
    
    font_texture_atlas::added_font_results_t font_texture_atlas::add_font(const font::font_file_t& file)
    {
        return add_font(file, file.character_min_index);
    }
    
    font_texture_atlas::added_font_results_t font_texture_atlas::add_font(const font::font_file_t& file, blt::u64 start)
    {
        blt::u32 max_height = 0;
        blt::u64 min_index_used = file.character_max_index;
        blt::u64 max_index_used = start;
        for (auto i = start; i < file.character_max_index; i++)
        {
            // returns an empty optional if there is no error code
            if (file.font.load_char(i))
                continue;
            auto face = *file.font.get();
            
            auto width = face->glyph->bitmap.width;
            auto height = face->glyph->bitmap.rows;
            max_height = std::max(max_height, height);
            
            // if adding this width overflows, we need to move to the next row.
            if (used_width + width > static_cast<blt::u32>(getWidth()))
            {
                used_height += max_height;
                max_height = 0;
                used_width = 0;
            }
            
            // if we can't fit the height of this character we should move to the next texture atlas.
            if (used_height + height > static_cast<blt::u32>(getHeight()))
                return {i, min_index_used, max_index_used};
            min_index_used = std::min(min_index_used, i);
            max_index_used = std::max(max_index_used, i);
            
            auto begin_width = used_width;
            auto begin_height = used_height;
            auto end_width = used_width + width;
            auto end_height = used_height + height;
            used_width += width;
            
            // upload the texture
            upload(face->glyph->bitmap.buffer, GL_RED, 0, static_cast<blt::i32>(begin_width), static_cast<blt::i32>(begin_height),
                   static_cast<blt::i32>(width), static_cast<blt::i32>(height));
            
            glyphs.insert(std::pair{i, bounded_t{
                    font::glyph_t{
                            {width, height},
                            {face->glyph->bitmap_left, face->glyph->bitmap_top},
                            face->glyph->advance.x},
                    {begin_width, begin_height}, {end_width, end_height}}
            });
        }
        return {-1ul, min_index_used, max_index_used};
    }
    
    font_generator_t::font_generator_t(blt::i32 dimensions, const std::vector<float>& sizes_to_generate):
            dimensions(dimensions), sizes_to_generate(sizes_to_generate)
    {}
    
    void font_generator_t::add_font(const font::font_file_t& file)
    {
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        float min_size = std::numeric_limits<float>::max();
        float max_size = 0;
        for (const auto& size : sizes_to_generate)
        {
            min_size = std::min(size, min_size);
            max_size = std::max(size, max_size);
            auto& vec = atlases;
            file.font.set_pixel_sizes(0, static_cast<blt::u32>(size));
            
            if (vec.empty())
                vec.emplace_back(std::make_unique<font_texture_atlas>(dimensions), min_size, max_size, file.character_min_index, 0);
            
            blt::u64 start_index = file.character_min_index;
            while (start_index != -1ul)
            {
                auto results = vec.back().atlas->add_font(file, start_index);
                vec.back().min_char = results.min_index_used;
                vec.back().max_char = results.max_index_used;
                vec.back().min_size = min_size;
                vec.back().max_size = max_size;
                min_size = size;
                max_size = size;
                if (results.last_inserted_index != -1ul)
                    vec.emplace_back(std::make_unique<font_texture_atlas>(dimensions), min_size, max_size, start_index, 0);
                start_index = results.last_inserted_index;
//                BLT_INFO("%f, %ld", size, start_index);
            }
        }
        glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
//        BLT_TRACE(atlases.size());
    }
    
    extern float square_vertices[20];
    extern const unsigned int square_indices[6];
    
    font_renderer_t::font_renderer_t()
    {}
    
    void font_renderer_t::create(const std::vector<float>& generated_font_sizes, blt::i32 dimensions)
    {
        font_shader = shader_t::make_unique(shader_2d_font_vert, shader_2d_font_frag);
        font::create();
        if (dimensions == 0)
        {
            int size;
            glGetIntegerv(GL_MAX_TEXTURE_SIZE, &size);
            dimensions = std::min(size, 4096);
            BLT_INFO("Max texture size %d, font renderer will use size %d", size, dimensions);
        }
        generator = std::make_unique<font_generator_t>(dimensions, generated_font_sizes);
    }
    
    void font_renderer_t::cleanup()
    {
        generator = nullptr;
        blt::gfx::font::cleanup();
    }
    
    font_renderer_t::compiled_text_t font_renderer_t::create_text(std::string_view str)
    {
        return font_renderer_t::compiled_text_t();
    }
    
}